/*
*  File             : runjob.c
*  Written by       : alinden@gmx.de
*  Copyright:       : GPL
*/

#include "conf.h"
#include "crond.h"
#include "extern.h"

void msgout(char *);
void log_info(char *);
void log_warn(char *);
struct passwd *namuid(char *);
void sig_dexit(int);
void runqproc(time_t, CONFIG *);
void restore_sigs(void);
void catf(FILE *, FILE *);
char *cpstr(size_t, char *, ...);
void getswap(void);

void stenv(CONFIG *tbuf, struct passwd *pword)   /* set the environment */
{
	NVR *nvrbuf;

	putenv(cpstr(1024, "HOME=%s", pword->pw_dir));
	putenv(cpstr(1024, "USER=%s", pword->pw_name));
	putenv(cpstr(1024, "LOGNAME=%s", pword->pw_name));
	putenv("SHELL=/bin/sh");
	putenv(DPATH);

	if(tbuf->jobid != NULL)
		putenv(cpstr(1024, "CRON_SUBJECT=%s", tbuf->jobid));

	for(nvrbuf = tbuf->envr; nvrbuf != NULL; nvrbuf = nvrbuf->nblock)
		putenv((char *)nvrbuf+sizeof(NVR));
}

/* Get the FQDN of the host. Thanks to Soenke Jan Peters <peters@simprovement.com> for
   mailing a patch */ 

char *getfqdn(void)
{
	char *fqdn,**h;
	struct hostent *hp;
	char addr[sizeof(struct in_addr)];

	fqdn = cpstr(0, NULL);
	fqdn[1024] = 0;

	/* Try gethostname() */
 
	if(gethostname(fqdn, 1024) == -1)
		return(NULL);

	if(strchr(fqdn, '.') != NULL)
		return(fqdn);
 
	/* Try gethostbyname() */

	if((hp = gethostbyname(fqdn)) == NULL)   
		return(NULL);

	if(strchr(hp->h_name, '.') != NULL)
	{
		strncpy(fqdn, hp->h_name, 1024);
		return(fqdn);
	}

	for(h = hp->h_aliases; *h != NULL; h++)
	{
		if(strchr(*h, '.') != NULL)
		{
			strncpy(fqdn, *h, 1024);
			return(fqdn);
		}	
	}

	/* Try gethostbyaddr() */
 
	memcpy(addr, (char*)hp->h_addr, sizeof(struct in_addr));

	if((hp = gethostbyaddr(addr, hp->h_length, hp->h_addrtype)) == NULL)
		return(NULL);

	if(strchr(hp->h_name, '.') != NULL)
	{
		strncpy(fqdn, hp->h_name, 1024);
		return(fqdn);
	}
 
	for(h = hp->h_aliases; *h != NULL; h++)
	{
		if(strchr(*h, '.') != NULL)
		{
			strncpy(fqdn, *h, 1024);
			return(fqdn);
		}
	}
	return(fqdn);
}

int set_user(struct passwd *pword, char *usernam)
{
	if(setgid(pword->pw_gid) < 0)
	{
		log_warn("setgid(): %m");
		return(0);
	}

	if(initgroups(usernam, pword->pw_gid) < 0)
	{
		log_warn("initgroups(): %m");
		return(0);
	}	

	if(setuid(pword->pw_uid) < 0)
	{
		log_warn("setuid(): %m");
		return(0);
	}
	return(1);
}	

void run_procs(CONFIG *tbuf)   /* start a job */
{
	struct passwd *pword;
	char *usernam,*tmpbuf,*ashell,*eshell,*ahome,*anice;
	int fodes;
	pid_t pid;

	if(logfile != NULL)
	{
		if(tbuf->jobid != NULL)
			msgout(cpstr(256, "User:%s Subject:%s", tbuf->username, tbuf->jobid)); 
		else
			msgout(cpstr(256, "User:%s Command:%s", tbuf->username, tbuf->cmdstr)); 
	}

	if(tbuf->lopt)
	{
		openlog("crond", LOG_PID, LOG_CRON);
		syslog(LOG_INFO, "%s", cpstr(42, "Start %s: %s", tbuf->username, tbuf->jobid?tbuf->jobid:tbuf->cmdstr));
		closelog();
	}

	tbuf->exite = 0;
	tbuf->exitc = 0;
	tbuf->swpdelay = 0;
	tbuf->qdelay = 0;
		
	cperiod = 1;

	usernam=cpstr(1024, "%s", tbuf->username);
	usernam[strcspn(usernam, "@")] = 0;

	if((pword = namuid(usernam)) == NULL)
		return;

	while((tmpbuf = tmpnam(NULL)) == NULL)
		sleep(1);

	while((tbuf->tmpbuf = malloc(strlen(tmpbuf)+1)) == NULL)
		sleep(1);

	strcpy(tbuf->tmpbuf, tmpbuf);

	while((pid = fork()) == -1)
		sleep(1);

	if(pid)
	{
		tbuf->pid = pid;
		time(&tbuf->lastexec);
		return;
	}	

	restore_sigs();
	
	if(svsock > -1)
		close(svsock);
	
	/* set the user id and environment */

	if(! set_user(pword, usernam))
		exit(1);
		
	stenv(tbuf, pword);

	/* we write the output from the command to a file in /tmp first 
	   to avoid mailbox locking while the program is running */
	
	if((fodes = open(tmpbuf, O_CREAT|O_WRONLY|O_EXCL, 0600)) != -1)
	{
		dup2(fodes, 2);
		dup2(2,1);
	}
	else
	{
		log_warn("cannot open temp file: %m");
		close(1);
		close(2);
	}

	if(tbuf->anice && (anice = getenv("NICE")) != NULL)
		nice(atoi(anice));
	else
		nice(CTABNICE); 

	if(tbuf->dopt || tbuf->period)
		signal(SIGTERM, (void *)sig_dexit);

	if(! tbuf->ahome || (ahome = getenv("HOME")) == NULL)
	{
		chdir(pword->pw_dir);
	}	
	else if(chdir(ahome) == -1)
	{
		perror(ahome);
		chdir(pword->pw_dir);
	}

	if(tbuf->ashell && (ashell = getenv("SHELL")) != NULL)
		eshell = ashell;
	else	
		eshell = "/bin/sh";

	execl(eshell, eshell, "-c", tbuf->cmdstr, NULL);
	perror(eshell);		
	exit(1);
}

void run_proc(CONFIG *tbuf)
{
	if(tbuf->atmout)
	{
		getswap();

		if(swaphigh)
		{
			if(! tbuf->swpdelay)
			{
				time(&tbuf->swpdelay);

				if(logfile != NULL)
					msgout(cpstr(256, "delay for %s", (tbuf->jobid)?(tbuf->jobid):(tbuf->cmdstr)));
			}

			if(time(NULL)-tbuf->swpdelay < tbuf->atmout)
			{
				tbuf->nextexec = time(NULL)+30;
				return;
			}
		}	
	}
	run_procs(tbuf);
}

void clear_job(CONFIG *jbuf)
{
	struct stat buf;
	FILE *mpipe,*tmpfile=NULL;
	char *mailto,*mailfrom,*fqdn;
	char *usernam;
	struct passwd *pword;
	pid_t pid;

	while((pid = fork()) == -1)
		sleep(1);

	if(pid)
		return;

	restore_sigs();

	usernam=cpstr(1024, "%s", jbuf->username);
	usernam[strcspn(usernam, "@")] = 0;

	if((pword = namuid(usernam)) == NULL)
		exit(1);
	
	if(jbuf->cmain && strchr(jbuf->username, '@') != NULL)
	{
		mailto = jbuf->username;
	}
	else if(jbuf->mailto != NULL)
	{
		mailto = jbuf->mailto;
	}
	else mailto = (jbuf->cmain)?("root"):(jbuf->username);

	if(*mailto && strcmp(mailto, "\"\""))
	{
		/* test if there was output from the job and drop a mail */

		if(stat(jbuf->tmpbuf, &buf) == 0 && buf.st_size > 0 && (tmpfile = fopen(jbuf->tmpbuf, "r")) != NULL && set_user(pword, usernam))
		{
			fqdn = getfqdn();
			mailfrom = (jbuf->cmain)?("root"):(pword->pw_name);

			if(fqdn != NULL)
				mailfrom = cpstr(1024, "%s@%s", mailfrom, fqdn);
		
			if((mpipe = popen(cpstr(1024, "%s %s", SENDMAIL, mailto), "w")) != NULL)
			{
				fprintf(mpipe, "From: CROND <%s>\nTo: %s\nSubject: %s\n\nOutput from command: %s\n\n", mailfrom,
				mailto, (jbuf->jobid != NULL)?(jbuf->jobid):("output from job"), jbuf->cmdstr);

				catf(tmpfile, mpipe);
				
				if(pclose(mpipe))
					log_warn(cpstr(1024, "mailer start returned error (%s)", SENDMAIL));
			}
			else
				log_warn("cannot run the mailer: %m");
		}
	}
	if(tmpfile != NULL)
		fclose(tmpfile);	

	unlink(jbuf->tmpbuf);

	exit(0);
}	

void mail_hmsg(CONFIG *jbuf)
{
	FILE *mpipe;
	char *mailto,*mailfrom,*fqdn;
	char *usernam;
	struct passwd *pword;
	pid_t pid;

	while((pid = fork()) == -1)
		sleep(1);

	if(pid)
		return;

	restore_sigs();

	usernam=cpstr(1024, "%s", jbuf->username);
	usernam[strcspn(usernam, "@")] = 0;

	if((pword = namuid(usernam)) == NULL)
		exit(1);
	
	if(jbuf->cmain && strchr(jbuf->username, '@') != NULL)
	{
		mailto = jbuf->username;
	}
	else if(jbuf->mailto != NULL && strcmp(jbuf->mailto, "\"\""))
	{
		mailto = jbuf->mailto;
	}
	else mailto = (jbuf->cmain)?("root"):(jbuf->username);

	if(*mailto && set_user(pword, usernam))
	{
		fqdn = getfqdn();
		mailfrom = (jbuf->cmain)?("root"):(pword->pw_name);

		if(fqdn != NULL)
			mailfrom = cpstr(1024, "%s@%s", mailfrom, fqdn);
	
		if((mpipe = popen(cpstr(1024, "%s %s", SENDMAIL, mailto), "w")) != NULL)
		{
			fprintf(mpipe, "From: CROND <%s>\nTo: %s\nSubject: %s\n\nJob exceeded %d minutes execution time\n\n%s", mailfrom,
			mailto, (jbuf->jobid != NULL)?(jbuf->jobid):("Time limit"), jbuf->htmout/60, jbuf->cmdstr);
			
			if(pclose(mpipe))
				log_warn(cpstr(1024, "mailer start returned error (%s)", SENDMAIL));
		}
		else
			log_warn("cannot run the mailer: %m");
	}
	exit(0);
}	

void rmpid(time_t ntmr)   /* clear terminated jobs */
{
	CONFIG *tbuf,*jbuf,*pbuf;
	pid_t cpid;
	int status;

	while((cpid = waitpid(-1, &status, WNOHANG)) > 0)   /* zombie */
	{
		for(tbuf = cblock; tbuf != NULL; tbuf = tbuf->nblock)
		{
			if(tbuf->pid == cpid)
			{
				tbuf->pid = tbuf->hsnd = 0;
				clear_job(tbuf);
				
				if(tbuf->tmpbuf != NULL)
				{
					free(tbuf->tmpbuf);
					tbuf->tmpbuf = NULL;
				}	
				
				if(WIFEXITED(status))
				{
					tbuf->exite = 0;
					tbuf->exitc = WEXITSTATUS(status);
				}	
				else tbuf->exite = 1;

				if(tbuf->qopt)  
					runqproc(ntmr, tbuf);
				
				break;
			}	
		}
		if(tbuf != NULL)
			continue;
			
		for(jbuf = jblock; jbuf != NULL; jbuf = jbuf->nblock)
		{
			if(jbuf->pid == cpid)
			{
				clear_job(jbuf);

				if(jbuf->tmpbuf != NULL)
					free(jbuf->tmpbuf);

				if(jblock == jbuf)
				{
					jblock = NULL;
				}
				else
				{
					for(pbuf = jblock; pbuf != NULL; pbuf = pbuf->nblock) 
					{
						if(pbuf->nblock == jbuf)
						{
							pbuf->nblock = jbuf->nblock;
							break;
						}
					}
				}	
				free(jbuf);
				
				break;
			}	
		} 
	}
} 
